(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de

 Module:	Colors

 Version:	20-Nov-1997

 Description:
	SetBitmapColors: Change entries of the DIB color table

	Some color system conversion stuff
	{ RGB <-> Hue, Saturation, Luminance (HSL) }

	How does Win95/NT 4 compute the	appropriate light and shadow
	colors of 3D-objects if the face color is given?
	I'm not sure, but Get3DColors obviously gives correct results
	thus proving my assumptions to be right.

 Copyright  1997  Robert Rossmair

 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind. If you use this software or portions
 of it in a commercial program, I would appreciate it if you mentioned my
 contribution in the documentation materials accompaning your work.
 ************************************************************************)
unit Colors;

interface

uses
  SysUtils,
  Windows,
  Graphics,
  Classes,
{$IFNDEF NoMath}
  Math,
{$ENDIF}
  Math32;

type
  THSLValue = 0..240;
  THSLVector = record
    Hue:	THSLValue;
    Saturation:	THSLValue;
    Luminance:	THSLValue;
  end;

{$IFNDEF NoMath}
procedure GammaCorrection(Gamma: Single; const Colors: array of TRGBQuad);
{$ENDIF}
procedure Get3DColors(FaceColor: TColor; var HighLightColor, ShadowColor: TColor);
function GetHSL(Color: TColor): THSLVector;
function GetHSLColor(H, S, L: Integer): TColor;
function HSL2RGB(const HSL: T3DVector): T3DVector;
function RGB2HSL(const RGB: T3DVector): T3DVector;
procedure GetBitmapColors(
  Bmp: TBitmap;
  var Colors: array of TColor;
  Index: Integer);
procedure SetBitmapColors(
  Bmp: TBitmap;
  Colors: array of TColor;
  Index: Integer);

implementation

uses Consts;

type
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array [Byte] of TRGBTriple;
  PRGBQuadArray = ^TRGBQuadArray;
  TRGBQuadArray = array [Byte] of TRGBQuad;
const
  Invalid = -1;

// taken from Graphics unit

procedure OutOfResources;
begin
  raise EOutOfResources.Create(SOutOfResources);
end;

procedure GDIError;
var
  ErrorCode: Integer;
  Buf: array [Byte] of Char;
begin
  ErrorCode := GetLastError;
  if (ErrorCode <> 0) and (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil,
    ErrorCode, LOCALE_USER_DEFAULT, Buf, sizeof(Buf), nil) <> 0) then
    raise EOutOfResources.Create(Buf)
  else
    OutOfResources;
end;

function GDICheck(Value: Integer): Integer;
begin
  if Value = 0 then GDIError;
  Result := Value;
end;

procedure GetBitmapColors(
  Bmp: TBitmap;
  var Colors: array of TColor;
  Index: Integer);
var
  i: Integer;
  ColorTable: PRGBQuadArray;
  Count: Integer;
begin
  Count := High(Colors)-Low(Colors)+1;
  GetMem(ColorTable, Count * SizeOf(TRGBQuad));
  try
    Bmp.HandleType := bmDIB;
    Count := GDICheck(GetDIBColorTable(Bmp.Canvas.Handle, Index, Count, ColorTable^));
    for i := 0 to Count-1 do
      with ColorTable^[i] do
	Colors[i-Index] := RGB(rgbRed, rgbGreen, rgbBlue);
  finally
    FreeMem(ColorTable);
  end;
end;

procedure SetBitmapColors(
  Bmp: TBitmap;
  Colors: array of TColor;
  Index: Integer);
var
  i, RGB: Integer;
  ColorTable: PRGBQuadArray;
  Count: Integer;
begin
  Count := High(Colors)-Low(Colors)+1;
  GetMem(ColorTable, Count * SizeOf(TRGBQuad));
  try
    for i := 0 to Count-1 do
      with ColorTable^[i] do
      begin
	RGB := ColorToRGB(Colors[i-Index]);
	rgbBlue		:= GetBValue(RGB);
	rgbGreen	:= GetGValue(RGB);
	rgbRed		:= GetRValue(RGB);
	rgbReserved	:= 0;
      end;
    Bmp.HandleType := bmDIB;
    GDICheck(SetDIBColorTable(Bmp.Canvas.Handle, Index, Count, ColorTable^));
  finally
    FreeMem(ColorTable);
  end;
end;
{$IFNDEF NoMath}
procedure GammaCorrection(Gamma: Single; const Colors: array of TRGBQuad);
const
  Inv255 = 1.0 / 255;
var
  i: Integer;
  InvGamma: double;
begin
  InvGamma := 1.0/Gamma;
  for i := Low(Colors) to High(Colors) do
  with Colors[i] do
  begin
    rgbRed	:= Round(255 * Power(rgbRed   * Inv255, InvGamma));
    rgbGreen	:= Round(255 * Power(rgbGreen * Inv255, InvGamma));
    rgbBlue	:= Round(255 * Power(rgbBlue  * Inv255, InvGamma));
  end;
end;
{$ENDIF}
function GetHSLColor(H, S, L: Integer): TColor;
var
  HSL: T3DVector;
  RGB: T3DVector;
begin
  HSL[0] := H/High(THSLValue);
  HSL[1] := S/High(THSLValue);
  HSL[2] := L/High(THSLValue);
  RGB := HSL2RGB(HSL);
  Result := Round(RGB[0] * 255)
	 or Round(RGB[1] * 255) shl 8
	 or Round(RGB[2] * 255) shl 16;
end;

function GetHSL(Color: TColor): THSLVector;
var
  HSL: T3DVector;
  RGB: T3DVector;
begin
  Color := ColorToRGB(Color);
  RGB[0] := GetRValue(Color)/255;
  RGB[1] := GetGValue(Color)/255;
  RGB[2] := GetBValue(Color)/255;
  HSL := RGB2HSL(RGB);
  with Result do
  begin
    Hue		:= Round(HSL[0] * High(THSLValue)) mod High(THSLValue);
    Saturation	:= Round(HSL[1] * High(THSLValue));
    Luminance	:= Round(HSL[2] * High(THSLValue));
  end;
end;

function HSL2RGB(const HSL: T3DVector): T3DVector;
const
  Hue: array[0..5, 0..2] of Integer = (
	(1, -1, -1),	// red
	(1, 1, -1),	// yellow
	(-1, 1, -1),	// green
	(-1, 1, 1),	// cyan
	(-1, -1, 1),	// blue
	(1, -1, 1));	// magenta
var
  i, j, k: Integer;
  x: Double;
begin
  j := Trunc(HSL[0] * 6) mod 6;
  k := (j+1) mod 6;
  x := Frac(HSL[0] * 6);
  for i := 0 to 2 do
    Result[i] := Hue[j, i] + x*(Hue[k, i]-Hue[j, i]);
  for i := 0 to 2 do Result[i] := Result[i] * HSL[1];
  if HSL[2] <= 0.5 then
    for i := 0 to 2 do Result[i] := HSL[2] * (Result[i]+ 1)
  else
    for i := 0 to 2 do Result[i] := HSL[2] + Result[i] * (1-HSL[2]);
end;

function RGB2HSL(const RGB: T3DVector): T3DVector;
const
  Epsilon = 1E-8;
var
  i, k: Integer;
  x: Double;
  V: T3DVector;
  W: T3DVector absolute Result;
  Hue: Double;
  Sat: Double;
  Lum: Double;

  function GetHue: Double;
  begin
    case k of
      0: if W[1] > W[2] then Result := 2+(W[2]+1)/2
	 else Result := 4-(W[1]+1)/2;
      1: if W[2] > W[0] then Result := 4+(W[0]+1)/2
	 else Result := 6-(W[2]+1)/2;
      2: if W[0] > W[1] then Result := (W[1]+1)/2
	 else Result := 2-(W[0]+1)/2;
      else Result := 0;
    end;
    Result := Result/6;
  end;

begin
  x := 0;
  for i := 0 to 2 do
  begin
    V[i] := 2*RGB[i]-1;
    if Abs(V[i]) > x then
    begin
      x := Abs(V[i]);
      k := i;
    end;
  end;
  try
    x := 1/x;
  except
    on EZeroDivide do
    begin
      Result[0] := 0;
      Result[1] := 0;
      Result[2] := 0.5;
      Exit;
    end;
  end;
  for i := 0 to 2 do W[i] := V[i] * x;
  x := 0;
  if V[k] <= 0 then
  begin
    for i := 0 to 2 do if (W[i]+1) > x then x := W[i] +1;
    try
      x := 2/x;
    except
      on EZeroDivide do	// RGB liegt auf der Grauachse
      begin
	Result[0] := 0;
	Result[1] := 0;
	Result[2] := (V[0]+1)*0.5;
	Exit;
      end;
    end;
    for i := 0 to 2 do W[i] := x*(W[i]+1)-1;
    Hue := GetHue;
    if Abs(V[1]-V[0]) > Epsilon then
      Sat := (V[1]-V[0])/(W[1]*(V[0]+1)-W[0]*(V[1]+1))
    else if Abs(V[2]-V[1]) > Epsilon then
      Sat := (V[2]-V[1])/(W[2]*(V[1]+1)-W[1]*(V[2]+1))
    else if Abs(V[2]-V[0]) > Epsilon then
      Sat := (V[2]-V[0])/(W[2]*(V[0]+1)-W[0]*(V[2]+1))
    else Sat := 0;

    if Abs(W[1]-W[0]) > Epsilon then
      Lum := (W[1]*(V[0]+1)-W[0]*(V[1]+1))/(W[1]-W[0])
    else if Abs(W[2]-W[1]) > Epsilon then
      Lum := (W[2]*(V[1]+1)-W[1]*(V[2]+1))/(W[2]-W[1])
    else if Abs(W[2]-W[0]) > Epsilon then
      Lum := (W[2]*(V[0]+1)-W[0]*(V[2]+1))/(W[2]-W[0])
    else Lum := V[0]+1;
    Lum := Lum * 0.5;
  end else
  begin
    for i := 0 to 2 do if (1-W[i]) > x then x := 1-W[i];
    try
      x := 2/x;
    except
      on EZeroDivide do	// RGB liegt auf der Grauachse
      begin
	Result[0] := 0;
	Result[1] := 0;
	Result[2] := (V[0]+1)*0.5;
	Exit;
      end;
    end;
    for i := 0 to 2 do W[i] := x*(W[i]-1)+1;
    x := 1;
    for i := 0 to 2 do
    if W[i] < x then
    begin
      x := W[i];
      k := i;
    end;
    Hue := GetHue;
    if Abs(V[1]-V[0]) > Epsilon then
      Sat := (V[1]-V[0])/(W[0]*(V[1]-1)-W[1]*(V[0]-1))
    else if Abs(V[2]-V[1]) > Epsilon then
      Sat := (V[2]-V[1])/(W[1]*(V[2]-1)-W[2]*(V[1]-1))
    else if Abs(V[2]-V[0]) > Epsilon then
      Sat := (V[2]-V[0])/(W[0]*(V[2]-1)-W[2]*(V[0]-1))
    else Sat := 0;

    if Abs(W[1]-W[0]) > Epsilon then
      Lum := (W[1]*(V[0]-1)-W[0]*(V[1]-1))/(W[1]-W[0])
    else if Abs(W[2]-W[1]) > Epsilon then
      Lum := (W[2]*(V[1]-1)-W[1]*(V[2]-1))/(W[2]-W[1])
    else if Abs(W[2]-W[0]) > Epsilon then
      Lum := (W[2]*(V[0]-1)-W[0]*(V[2]-1))/(W[2]-W[0])
    else Lum := V[0]-1;
    Lum := 1 + Lum * 0.5;
  end;
  W[0] := Hue;
  W[1] := Sat;
  W[2] := Lum;
end;

procedure Get3DColors(FaceColor: TColor; var HighLightColor, ShadowColor: TColor);
var
  V, HSL: T3DVector;
begin
  FaceColor := ColorToRGB(FaceColor);
  HighLightColor := RGB(
	255-(256-GetRValue(FaceColor)) div 2,
	255-(256-GetGValue(FaceColor)) div 2,
	255-(256-GetBValue(FaceColor)) div 2);
  V[0] := GetRValue(FaceColor)/255;
  V[1] := GetGValue(FaceColor)/255;
  V[2] := GetBValue(FaceColor)/255;
  HSL := RGB2HSL(V);
  HSL[2] := HSL[2] * (2/3);	// Luminance
  V := HSL2RGB(HSL);
  ShadowColor := RGB(Round(V[0]*255), Round(V[1]*255), Round(V[2]*255));
end;

end.
